"""Linked list utilities.

The code for this module is lifted from
`here <https://www.educative.io/edpresso/how-to-merge-two-sorted-linked-lists-in-python>`_.
"""

from __future__ import annotations

from typing import Any

from attrs import define


@define
class Node:
    """Node for the linked list.

    Parameters
    ----------
    data : Any, optional (default None)
        The current data.
    next : Node | None, optional (default None)
        The next data in the chain.
    """

    data: Any = None
    next: Node | None = None

    def to_list(self) -> list[Any]:
        """Convert the nodes to a de-duped list.

        Returns
        -------
        list[Any]
            A standard list of data.
        """
        out: list[Any] = []
        while self.next:
            out.append(self.data)
            self = self.next
        out.append(self.data)

        return out


@define
class LinkedList:
    """The linked list.

    Parameters
    ----------
    head : Node, optional (default None)
        The start of the list.
    """

    head: Node | None = None

    def append(self, data: Any) -> None:
        """Append a new node to the end of the list.

        Parameters
        ----------
        data : Any
            The new data.
        """
        new = Node(data=data)
        if self.head:
            # Scroll to the end of the list
            current = self.head
            while current.next:
                current = current.next
            current.next = new
        else:
            self.head = new

    @staticmethod
    def from_list(data: list[Any]) -> LinkedList:
        """Convert a standard list to a linked list."""
        # Sort the list
        sorted_list = sorted(data)
        out = LinkedList()
        for val in sorted_list:
            out.append(val)

        return out


def merge(list1: Node, list2: Node) -> Node:
    """Merge two linked lists.

    Parameters
    ----------
    list1 : Node
        The head of the first list.
    list2 : Node
        The head of the second list.

    Returns
    -------
    Node
        The head of the final, merged list.
    """
    # Create the head of the new list as a dummy node
    head = Node()
    current = head
    # Loop while the data exists
    while list1 and list2:
        # Add the earliest available node from either l1 or l2
        if list1.data < list2.data:
            current.next = list1
            # Scroll the list
            list1: Node | None = list1.next  # type: ignore[no-redef]
        else:
            current.next = list2
            # Scroll the list
            list2: Node | None = list2.next  # type: ignore[no-redef]
        current = current.next
    # Add non-empty list
    current.next = list1 or list2
    return head.next  # type: ignore[return-value]



"""Util methods."""

import inspect
import json
from collections.abc import Iterator
from datetime import datetime, timezone
from typing import Any

from attrs import Attribute, asdict, fields, filters

from lazyscribe.artifacts.base import Artifact
from lazyscribe.exception import ArtifactLoadError
from lazyscribe.registry import registry


def serializer(inst: type, field: "Attribute[Any]", value: Any) -> Any:
    """Datetime and dependencies converter for :meth:`attrs.asdict`.

    Parameters
    ----------
    inst : type
        Included for compatibility.
    field : attrs.Attribute[Any]
        The field name.
    value : Any
        The field value.

    Returns
    -------
    Any
        Converted value for easy serialization.
    """
    if isinstance(value, datetime):
        return value.isoformat(timespec="seconds")
    if field is not None and field.name == "dependencies":
        deps: list[str] = []
        for exp in value.values():
            if (project := registry.search(exp.project)) is not None:
                deps.append(f"{project}|{exp.slug}")
            else:
                deps.append(f"{exp.project}|{exp.slug}")
        return deps
    if field is not None and field.name == "tests":
        tests: list[dict[str, Any]] = [asdict(test) for test in value]
        return tests
    if field is not None and field.name == "artifacts":
        art: list[dict[str, Any]] = list(serialize_artifacts(value))
        return art

    return value


def serialize_artifacts(alist: list[Artifact]) -> Iterator[dict[str, Any]]:
    """Serialize list of artifacts."""
    yield from (
        {
            **asdict(
                artifact,
                filter=filters.exclude(
                    fields(type(artifact)).value,
                    fields(type(artifact)).writer_kwargs,
                    fields(type(artifact)).dirty,
                ),
                value_serializer=lambda _, __, value: value.isoformat(
                    timespec="seconds"
                )
                if isinstance(value, datetime)
                else value,
            ),
            "handler": artifact.alias,
        }
        for artifact in alist
    )


def utcnow() -> datetime:
    """Return the naive datetime now in UTC.

    Returns
    -------
    datetime.datetime
        Now in UTC, without timezone info.
    """
    return datetime.now(timezone.utc).replace(tzinfo=None)


def validate_artifact_environment(artifact: Artifact) -> None:
    """Validate the artifact handler environment.

    Parameters
    ----------
    artifact : Artifact
        An artifact handler instantiated from project and/or repository metadata.

    Raises
    ------
    lazyscribe.exception.ArtifactLoadError
        Raised if the runtime environment does not match artifact metadata.
    """
    # Construct the handler with relevant parameters.
    artifact_attrs: dict[str, Any] = {
        x: y
        for x, y in inspect.getmembers(artifact)
        if not x.startswith("_") and not inspect.ismethod(y)
    }
    # Exclude parameters that don't define equality
    exclude_names: list[str] = [
        attr.name for attr in fields(type(artifact)) if not attr.eq
    ]
    construct_params: list[str] = [
        param_name
        for param_name, param in inspect.signature(
            artifact.construct
        ).parameters.items()
        if param_name not in exclude_names or param.default == param.empty
    ]
    artifact_attrs = {
        key: value for key, value in artifact_attrs.items() if key in construct_params
    }

    curr_handler = type(artifact).construct(**artifact_attrs, dirty=False)
    # Validate the handler
    if curr_handler != artifact:
        field_filters = filters.exclude(
            *[attr for attr in fields(type(artifact)) if not attr.eq]
        )
        raise ArtifactLoadError(
            "Runtime environments do not match. Artifact parameters:\n\n"
            f"{json.dumps(asdict(artifact, filter=field_filters))}"
            "\n\nCurrent parameters:\n\n"
            f"{json.dumps(asdict(curr_handler, filter=field_filters))}"
        )



"""Main module."""

from lazyscribe._meta import __version__  # noqa: F401
from lazyscribe.experiment import Experiment
from lazyscribe.project import Project
from lazyscribe.repository import Repository
from lazyscribe.test import Test

__all__: list[str] = ["Experiment", "Project", "Repository", "Test"]



"""Repository releases."""

from __future__ import annotations

import json
import logging
import sys
from bisect import bisect
from datetime import datetime
from io import IOBase
from operator import attrgetter
from pathlib import Path
from typing import Any, Literal

from attrs import Factory, define, field

from lazyscribe._utils import utcnow
from lazyscribe.exception import VersionNotFoundError
from lazyscribe.repository import Repository

# Conditional import of the tomli library
if sys.version_info < (3, 11):
    import tomli
else:
    import tomllib as tomli

LOG = logging.getLogger(__name__)


@define
class Release:
    """Create a release associated with a Repository instance.

    Parameters
    ----------
    tag : str
        A string descriptor for the release. Commonly coincides with semantic
        or calendar versioning.
    artifacts : list[tuple[str, int]]
        A list of the latest available artifacts and versions in the source repository.
    created_at : datetime.datetime, optional (default ``lazyscribe._utils.utcnow()``)
        The creation timestamp for the release (in UTC).
    """

    tag: str = field()
    artifacts: list[tuple[str, int]] = field()
    created_at: datetime = Factory(utcnow)

    def to_dict(self) -> dict[str, list[tuple[str, int]] | str]:
        """Serialize the release to a dictionary.

        Returns
        -------
        dict
            A dictionary with the release information.
        """
        return {
            "tag": self.tag,
            "artifacts": self.artifacts,
            "created_at": self.created_at.isoformat(timespec="seconds"),
        }

    @classmethod
    def from_dict(cls, info: dict[str, Any]) -> Release:
        """Convert a serialized representation of the release back to a python object.

        Parameters
        ----------
        info : dict
            The dictionary representation of the release.

        Returns
        -------
        lazyscribe.release.Release
            The new release object.
        """
        return cls(
            tag=info["tag"],
            artifacts=info["artifacts"],
            created_at=datetime.fromisoformat(info["created_at"]),
        )


def create_release(repository: Repository, tag: str) -> Release:
    """Create a release.

    A release is a collection of specific artifact versions. It is generated
    by taking the latest available version of each artifact.

    Parameters
    ----------
    repository : lazyscribe.repository.Repository
        The source repository.
    tag : str
        A string descriptor of the tag. Commonly coincides with semantic or calendar
        versioning.

    Returns
    -------
    lazyscribe.release.Release
        The release object.
    """
    if repository.mode != "r":
        raise RuntimeError("Repository must be in read-only mode for filtering.")
    if any(art.dirty for art in repository.artifacts):
        raise RuntimeError(
            "At least one artifact has changed since it was last saved. Please save your "
            "repository and re-open it in read-only mode before filtering."
        )

    all_artifacts_ = sorted({art.name for art in repository.artifacts})
    latest: list[tuple[str, int]] = []
    for name in all_artifacts_:
        art = repository._search_artifact_versions(name)
        latest.append((name, art.version))

    return Release(tag=tag, artifacts=latest)


def find_release(
    releases: list[Release],
    version: str | datetime | None = None,
    match: Literal["asof", "exact"] = "exact",
) -> Release:
    """Find a release based on tag or timestamp.

    Parameters
    ----------
    releases : list[lazyscribe.release.Release]
        The releases associated with the repository.
    version : str | datetime, optional (default None)
        The version to find. If a string is provided, the function will assume
        the value corresponds to a ``tag``. If a datetime is provided, the function
        will assume the value corresponds to a creation date. If None is provided,
        the latest release will be returned.
    match : {"asof", "exact"}, optional (default "exact")
        Matching logic. Only relevant if a datetime is provided for ``version``.
        ``exact`` will provide the release with the exact value matching ``version``.
        ``asof`` will provide the most recent release as of the ``version`` datetime
        provided.

    Returns
    -------
    lazyscribe.release.Release
        The release object.

    Raises
    ------
    lazyscribe.exception.VersionNotFoundError
        Raised if a release cannot be found.
    ValueError
        Raised if the specified matching logic does not match the version type specified.
    """
    out: Release

    releases_ = sorted(releases, key=lambda x: x.created_at)
    match (match, version):
        case (_, None):
            out = releases_[-1]
        case ("exact", str() as tag):
            try:
                out = next(ver for ver in releases_ if ver.tag == tag)
            except StopIteration:
                msg = f"Cannot find release with tag '{tag}'"
                raise VersionNotFoundError(msg) from None
        case ("exact", datetime() as created_at):
            try:
                out = next(ver for ver in releases_ if ver.created_at == created_at)
            except StopIteration:
                msg = f"Cannot find release with creation date of {created_at!s}"
                raise VersionNotFoundError(msg) from None
        case ("asof", str()):
            raise ValueError(
                "Cannot perform an ``asof`` match using a tag. Please provide a "
                "``datetime.datetime`` value to perform an ``asof`` match."
            )
        case ("asof", datetime() as created_at):
            eligible_idx_ = bisect(releases_, created_at, key=attrgetter("created_at"))
            eligible_releases_ = releases_[:eligible_idx_]

            if len(eligible_releases_) == 0:
                msg = (
                    f"Creation date {created_at!s} predates the earliest release: "
                    f"'{releases_[0].tag}' ({releases_[0].created_at!s})"
                )
                raise VersionNotFoundError(msg)

            out = eligible_releases_[-1]
        case _:
            msg = "Invalid parameterization. Please provide ``exact`` or ``asof`` for ``match``."
            raise ValueError(msg)

    LOG.info(f"Found release '{out.tag}' ({out.created_at!s})")

    return out


def dump(obj: list[Release], fp: IOBase, **kwargs: Any) -> None:
    """Write the releases data.

    .. code-block:: python

        from lazyscribe import release as lzr

        releases: list[lazyscribe.release.Release]
        with open("releases.json", "w") as outfile:
            lzr.dump(releases, outfile)

    Parameters
    ----------
    obj : list[lazyscribe.release.Release]
        The list of release objects.
    fp : io.IOBase
        A buffer we can write to.
    **kwargs
        Keyword arguments for ``json.dump``.
    """
    json.dump([ver.to_dict() for ver in obj], fp, **kwargs)


def dumps(obj: list[Release], **kwargs: Any) -> str:
    """Convert a list of releases to a JSON-serialized string.

    To prevent namespace confusion, we recommend importing this function
    through an alias:

    .. code-block:: python

        from lazyscribe import release as lzr

        releases: list[lzr.Release]
        out = lzr.dumps(releases)

    Parameters
    ----------
    obj : list[lazyscribe.release.Release]
        The list of release objects.
    **kwargs
        Keyword arguments for ``json.dumps``.

    Returns
    -------
    str
        The JSON-serialized string.
    """
    return json.dumps([ver.to_dict() for ver in obj], **kwargs)


def load(fp: IOBase, **kwargs: Any) -> list[Release]:
    """Generate a list of releases from a file buffer.

    To prevent namespace confusion, we recommend importing this function
    through an alias:

    .. code-block:: python

        from lazyscribe import release as lzr

        with open("releases.json") as infile:
            releases = lzr.load(infile)

    Parameters
    ----------
    fp : file-like object
        A buffer that we can read using JSON.
    **kwargs
        Keyword arguments for ``json.load``

    Returns
    -------
    list[lazyscribe.release.Release]
        A list of releases.
    """
    json_data_ = json.load(fp, **kwargs)

    return [Release.from_dict(ver) for ver in json_data_]


def loads(s: str, **kwargs: Any) -> list[Release]:
    """Generate a list of releases from a string.

    To prevent namespace confusion, we recommend importing this module
    through an alias:

    .. code-block:: python

        from lazyscribe import release as lzr

        mydata = '[{"tag": "v0.1.0", "artifacts": [], "created_at": "2025-01-01T00:00:00"}]'
        releases = lzr.loads(mydata)

    Parameters
    ----------
    s : str
        The string representation of a JSON file.
    **kwargs
        Keyword arguments for ``json.loads``

    Returns
    -------
    list[lazyscribe.release.Release]
        A list of releases.
    """
    json_data_ = json.loads(s, **kwargs)

    return [Release.from_dict(ver) for ver in json_data_]


def release_from_toml(cfg: str) -> None:
    """Generate a release for supplied repositories from a configuration.

    This function will read in a TOML-compatible configuration file and look for
    the ``[tool.lazyscribe]`` table. This table must contain 1 field:

    * ``repositories`` (list): path to repository JSON files for which we want releases.

    The configuration has optional fields, including

    * ``version``: the current version of the overall project. If not supplied,
      this function will look for the ``version`` attribute of the ``[project]`` table.
    * ``format``: format for the repository release versions. This string will be formatted with the
      ``version`` string, as well as the ``year``, ``month``, and ``day`` of the release. By default,
      this format is ``v{version}``.

    This function will read in each repository, create a new release, and write it to a ``releases.json``
    file in the same directory as the source repository JSON file.

    Suppose you have the following entry in ``pyproject.toml``:

    .. code-block:: toml

        [project]
        version = "1.0.0"

        ...

        [tool.lazyscribe]
        repositories = [
            "src/models/model-1/repository.json",
            "src/models/model-2/repository.json"
        ]

    calling

    .. code-block:: python

        import lazyscribe.release as lzr

        with open("pyproject.toml") as infile:
            release_from_toml(infile.read())

    will create two new files:

    * ``src/models/model-1/releases.json``, and
    * ``src/models/model-2/release.json``.

    Each of these files will contain a ``v1.0.0`` release.

    Parameters
    ----------
    cfg : str
        String contents of the configuration file
    """
    cfg_data_ = tomli.loads(cfg)

    try:
        curr_version_ = cfg_data_["project"]["version"]
    except KeyError:
        curr_version_ = cfg_data_["tool"]["lazyscribe"]["version"]

    version_format_ = cfg_data_["tool"]["lazyscribe"].get("format", "v{version}")

    repositories = cfg_data_["tool"]["lazyscribe"]["repositories"]
    for fpath in repositories:
        repo = Repository(fpath, mode="r")
        new_release_ = create_release(repo, tag="__placeholder")
        new_release_.tag = version_format_.format(
            version=curr_version_,
            year=new_release_.created_at.year,
            month=new_release_.created_at.month,
            day=new_release_.created_at.day,
        )
        # Read in current releases
        release_fpath = Path(fpath).parent / "releases.json"
        if release_fpath.exists():
            with open(release_fpath) as infile:
                curr_releases_ = load(infile)

            # Check if the release already exists
            try:
                _ = find_release(
                    curr_releases_, version=new_release_.tag, match="exact"
                )
            except VersionNotFoundError:
                curr_releases_.append(new_release_)
                with open(release_fpath, "w") as outfile:
                    dump(curr_releases_, outfile, indent=4)
                continue
            LOG.warning(
                "Release '%s' already exists for the repository at '%s'. Skipping...",
                new_release_.tag,
                fpath,
            )
        else:
            with open(release_fpath, "w") as outfile:
                dump([new_release_], outfile, indent=4)



"""Sub-population tests."""

from __future__ import annotations

from typing import Any

from attrs import Factory, asdict, define, field, frozen

from lazyscribe._utils import serializer


@define
class Test:
    """Sub-population tests.

    These objects should only be instantiated within an experiment. A test is associated with
    some subset of the entire experiment. For example, a test could be used to evaluate the
    performance of a model against a specific subpopulation.

    Parameters
    ----------
    name : str
        The name of the test.
    description : str, optional (default None)
        A description of the test.
    metrics : dict[str, float | int], optional (default {})
        A dictionary of metric values. Each metric value can be an individual value or a list.
    parameters : dict[str, Any], optional (default {})
        A dictionary of test parameters. The key must be a string but the value can be anything.
    """

    # Tell pytest it's not a Python test class
    __test__ = False

    name: str = field()
    description: str | None = Factory(lambda: None)
    metrics: dict[str, float | int] = Factory(lambda: {})
    parameters: dict[str, Any] = Factory(lambda: {})

    def log_metric(self, name: str, value: float | int) -> None:
        """Log a metric to the test.

        This method will overwrite existing keys.

        Parameters
        ----------
        name : str
            Name of the metric.
        value : int | float
            Value of the metric.
        """
        self.metrics[name] = value

    def __str__(self) -> str:
        """Shortened string representation."""
        return f"<lazyscribe.test.Test at {hex(id(self))}>"

    def log_parameter(self, name: str, value: Any) -> None:
        """Log a parameter to the test.

        This method will overwrite existing keys.

        Parameters
        ----------
        name : str
            The name of the parameter.
        value : Any
            The parameter itself.
        """
        self.parameters[name] = value

    def to_dict(self) -> dict[str, Any]:
        """Serialize the test to a dictionary.

        Returns
        -------
        dict[str, Any]
            The test dictionary.
        """
        return asdict(
            self,
            value_serializer=serializer,
        )


@frozen
class ReadOnlyTest(Test):
    """Immutable version of the test."""

    def __str__(self) -> str:
        """Shortened string representation."""
        return f"<lazyscribe.test.ReadOnlyTest at {hex(id(self))}>"



"""Base class for new artifact handlers."""

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from datetime import datetime
from io import IOBase
from typing import Any, ClassVar

from attrs import define, field


@define
class Artifact(metaclass=ABCMeta):
    """Generic artifact handler that defines the expected interface.

    Artifact handlers are not meant to be initialized directly.

    Attributes
    ----------
    alias : str
        The alias for the artifact handler. This value will be supplied to
        :py:meth:`lazyscribe.experiment.Experiment.log_artifact`.
        (A class attribute.)
    suffix : str
        The standard suffix for the files written and read by this handler.
        (A class attribute.)
    binary : bool
        Whether or not the file format for the handler is binary in nature. This
        affects whether or not the file handler uses ``w`` or ``wb``.
        (A class attribute.)
    output_only : bool
        Whether or not the file output by the handler is meant to be read as the orginal project.
        (A class attribute.)
    name : str
        The name of the artifact.
    fname : str
        The filename of the artifact.
    value : Any
        The value for the artifact.
    writer_kwargs : dict
        User provided keyword arguments for writing an artifact. Provided when
        the artifact is logged to an experiment.
    version : int
        Version of the artifact.
    dirty : bool
        Whether or not this artifact should be saved when :py:meth:`lazyscribe.project.Project.save`
        or :py:meth:`lazyscribe.repository.Repository.save` is called. This decision is based
        on whether the artifact is new or has been updated.
    """

    alias: ClassVar[str]
    suffix: ClassVar[str]
    binary: ClassVar[bool]
    output_only: ClassVar[
        bool
    ]  # Describes if the artifact will reconstruct to a Python object on read
    name: str = field(eq=False)
    fname: str = field(eq=False)
    value: Any = field(eq=False)
    writer_kwargs: dict[str, Any] = field(eq=False)
    created_at: datetime = field(eq=False)
    expiry: datetime | None = field(eq=False)
    version: int = field(eq=False)
    dirty: bool = field(eq=False)

    @classmethod
    @abstractmethod
    def construct(
        cls,
        name: str,
        value: Any = None,
        fname: str | None = None,
        created_at: datetime | None = None,
        expiry: datetime | None = None,
        writer_kwargs: dict[str, Any] | None = None,
        version: int = 0,
        dirty: bool = True,
        **kwargs: Any,
    ) -> Artifact:
        """Construct the artifact handler.

        This method should use environment variables to capture information that
        is relevant to compatibility between runtime environments.

        Parameters
        ----------
        name : str
            The name of the artifact.
        value : Any, optional (default None)
            The value for the artifact.
        fname : str, optional (default None)
            The filename for the artifact. If set to ``None`` or not provided, it will be derived from
            the name of the artifact and the suffix for the class.
        created_at : datetime.datetime, optional (default ``lazyscribe._utils.utcnow()``)
            When the artifact was created.
        expiry : datetime.datetime, optional (default None)
            When the artifact expired.
        writer_kwargs : dict, optional (default {})
            Keyword arguments for writing an artifact to the filesystem. Provided when an artifact
            is logged to an experiment.
        version : int, optional (default 0)
            Integer version to be used for versioning artifacts.
        dirty : bool, optional (default True)
            Whether or not this artifact should be saved when :py:meth:`lazyscribe.project.Project.save`
            or :py:meth:`lazyscribe.repository.Repository.save` is called. This decision is based
            on whether the artifact is new or has been updated.
        **kwargs
            Other keyword arguments.

        Returns
        -------
        Artifact
            The artifact.
        """

    @classmethod
    @abstractmethod
    def read(cls, buf: IOBase, **kwargs: Any) -> Any:
        """Read in the artifact.

        Parameters
        ----------
        buf : file-like object
            The buffer from a ``fsspec`` filesystem.
        **kwargs
            Keyword arguments for the read method.

        Returns
        -------
        Any
            The artifact object.
        """

    @classmethod
    @abstractmethod
    def write(cls, obj: Any, buf: IOBase, **kwargs: Any) -> None:
        """Write the artifact to the filesystem.

        Parameters
        ----------
        obj : Any
            The object to write to the buffer.
        buf : file-like object
            The buffer from a ``fsspec`` filesystem.
        **kwargs
            Keyword arguments for the write method.
        """



"""Experiment dataclass."""

from __future__ import annotations

import getpass
import logging
import os
import warnings
from collections.abc import Iterator
from contextlib import contextmanager
from copy import copy
from datetime import datetime
from pathlib import Path
from typing import Any

from attrs import (
    Factory,
    asdict,
    define,
    evolve,
    field,
    fields,
    filters,
    frozen,
)
from fsspec.implementations.local import LocalFileSystem
from fsspec.spec import AbstractFileSystem
from slugify import slugify

from lazyscribe._utils import serializer, utcnow, validate_artifact_environment
from lazyscribe.artifacts import _get_handler
from lazyscribe.artifacts.base import Artifact
from lazyscribe.exception import ArtifactLoadError, ArtifactLogError, SaveError
from lazyscribe.repository import Repository
from lazyscribe.test import Test

LOG = logging.getLogger(__name__)


@define
class Experiment:
    """Experiment data class.

    This class is not meant to be initialized directly. It is meant to be used through the
    :py:class:`lazyscribe.project.Project` class.

    Parameters
    ----------
    name : str
        The name of the experiment.
    project : pathlib.Path
        The path to the project JSON associated with the project.
    dir : pathlib.Path, optional (default None)
        Directory for the project and the experiment. If not supplied, the parent directory for the ``project`` file
        will be used.
    author : str, optional (default ``getpass.getuser()``)
        The author of the experiment.
    last_updated_by : str, optional (default None)
        Last editor of the experiment. If not supplied, the ``author`` will be used.
    metrics : dict[str, float | int], optional (default {})
        A dictionary of metric values. Each metric value can be an individual value or a list.
    parameters : dict[str, Any], optional (default {})
        A dictionary of experiment parameters. The key must be a string but the value can be
        anything.
    created_at : datetime.datetime, optional (default ``lazyscribe._utils.utcnow()``)
        When the experiment was created (in UTC).
    last_updated : datetime.datetime, optional (default ``lazyscribe._utils.utcnow()``)
        When the experiment was last updated (in UTC).
    short_slug : str, optional (default None)
        Slugified ``name``. Defaults to calling :py:meth:`slugify.slugify` on the ``name`` attribute.
    slug : str, optional (default None)
        Unique identifier for the experiment. Defaults to the slugified ``name`` with the creation date
        appended in the format ``YYYYMMDDHHMMSS``.
    tags : list[str], optional (default [])
        Tags for filtering and identifying experiments across a project.
    dependencies : dict[str, lazyscribe.experiment.Experiment], optional (default {})
        A dictionary of upstream project experiments. The key is the short slug for the upstream
        experiment and the value is an :py:class:`Experiment` instance.
    tests : list[lazyscribe.test.Test], optional (default [])
        List of :py:class:`lazyscribe.test.Test` objects corresponding to sub-population/non-global metrics.
    artifacts : list[lazyscribe.artifacts.base.Artifact], optional (default [])
        List of :py:class:`lazyscribe.artifact.base.Artifact` objects corresponding to experimental artifacts.
    dirty : bool, optional (default True)
        Whether or not this experiment should be saved when :py:meth:`lazyscribe.project.Project.save`
        is called. This decision is based on whether the experiment is new or has been updated.
    """

    name: str = field()
    project: Path = field(eq=False)
    dir: Path = field(eq=False)
    fs: AbstractFileSystem = field(eq=False)
    author: str = Factory(getpass.getuser)
    last_updated_by: str = field()
    metrics: dict[str, float | int] = Factory(lambda: {})
    parameters: dict[str, Any] = Factory(lambda: {})
    created_at: datetime = Factory(utcnow)
    last_updated: datetime = Factory(utcnow)
    dependencies: dict[str, Experiment] = field(eq=False, factory=lambda: {})
    short_slug: str = field()
    slug: str = field()
    tests: list[Test] = Factory(lambda: [])
    artifacts: list[Artifact] = Factory(factory=lambda: [])
    tags: list[str] = Factory(factory=lambda: [])
    dirty: bool = field(eq=False, factory=lambda: True)

    @dir.default
    def _dir_factory(self) -> Path:
        """Get the default directory for the project and the experiment.

        Returns
        -------
        pathlib.Path
            Absolute path to the directory.
        """
        return self.project.parent

    @fs.default
    def _fs_default(self) -> AbstractFileSystem:
        """Define a default local filesystem implementation.

        Returns
        -------
        fsspec.implementations.local.LocalFileSystem
            A standard local filesystem through ``fsspec``.
        """
        return LocalFileSystem()

    @last_updated_by.default
    def _last_updated_by_factory(self) -> str:
        """Last editor."""
        return self.author

    @short_slug.default
    def _short_slug_factory(self) -> str:
        """Get the short slug.

        Returns
        -------
        str
            The slugified experiment name.
        """
        return slugify(self.name)

    @slug.default
    def _slug_factory(self) -> str:
        """Get the full experiment slug.

        Returns
        -------
        str
            Experiment slug, in the format `{name}-{created_at}`.
        """
        return slugify(f"{self.name}-{self.created_at.strftime('%Y%m%d%H%M%S')}")

    @property
    def path(self) -> Path:
        """Path to an experiment folder.

        This folder can be used to store any plots or artifacts that you want to associate
        with the experiment.

        Returns
        -------
        pathlib.Path
            The path for the experiment.
        """
        return self.dir / self.slug

    def log_metric(self, name: str, value: float | int) -> None:
        """Log a metric to the experiment.

        This method will overwrite existing keys.

        Parameters
        ----------
        name : str
            Name of the metric.
        value : int | float
            Value of the metric.
        """
        self.last_updated = utcnow()
        self.metrics[name] = value

        self.dirty = True

    def log_parameter(self, name: str, value: Any) -> None:
        """Log a parameter to the experiment.

        This method will overwrite existing keys.

        Parameters
        ----------
        name : str
            The name of the parameter.
        value : Any
            The parameter itself.
        """
        self.last_updated = utcnow()
        self.parameters[name] = value

        self.dirty = True

    def tag(self, *args: str, overwrite: bool = False) -> None:
        """Add one or more tags to the experiment.

        .. important::

            If this function is called with no supplied values for ``*args``
            _and_ ``overwrite=True``, the result will be that the experiment has
            no associated tags.

        Parameters
        ----------
        *args
            The tags.
        overwrite : bool, optional (default False)
            Whether to add or overwrite the new tags.
        """
        self.last_updated = utcnow()
        new_tags_ = list(args)
        if overwrite:
            self.tags = new_tags_
        else:
            self.tags += new_tags_

        self.dirty = True

    def log_artifact(
        self,
        name: str,
        value: Any,
        handler: str,
        fname: str | None = None,
        overwrite: bool = False,
        **kwargs: Any,
    ) -> None:
        """Log an artifact to the experiment.

        This method associates an artifact with the experiment, but the artifact will
        not be written until :py:meth:`lazyscribe.Project.save` is called.

        Parameters
        ----------
        name : str
            The name of the artifact.
        value : Any
            The object to persist to the filesystem.
        handler : str
            The name of the handler to use for the object.
        fname : str, optional (default None)
            The filename for the artifact. If not provided, it will be derived from the
            name of the artifact and the builtin suffix for each handler.
        overwrite : bool, optional (default False)
            Whether or not to overwrite an existing artifact with the same name. If set to ``True``,
            the previous artifact will be removed and overwritten with the current artifact.
        **kwargs
            Keyword arguments for the write function of the handler.

        Raises
        ------
        lazyscribe.exception.ArtifactLogError
            Raised if an artifact is supplied with the same name as an existing artifact and
            ``overwrite`` is set to ``False``.
        """
        # Retrieve and construct the handler
        self.last_updated = utcnow()
        self.dirty = True
        handler_cls = _get_handler(handler)
        artifact_handler = handler_cls.construct(
            name=name,
            value=value,
            fname=fname,
            created_at=self.last_updated,
            writer_kwargs=kwargs,
        )
        for index, artifact in enumerate(self.artifacts):
            if artifact.name == name:
                if overwrite:
                    self.artifacts[index] = artifact_handler
                    if handler_cls.output_only:
                        warnings.warn(
                            f"Artifact '{name}' is added. It is not meant to be read back as Python Object",
                            UserWarning,
                            stacklevel=2,
                        )
                    break
                else:
                    raise ArtifactLogError(
                        f"An artifact with name {name} already exists in the experiment. Please "
                        "use another name or set ``overwrite=True`` to replace the artifact."
                    )
        else:
            self.artifacts.append(artifact_handler)
            if handler_cls.output_only:
                warnings.warn(
                    f"Artifact '{name}' is added. It is not meant to be read back as Python Object",
                    UserWarning,
                    stacklevel=2,
                )

    def load_artifact(self, name: str, validate: bool = True, **kwargs: Any) -> Any:
        """Load a single artifact.

        Parameters
        ----------
        name : str
            The name of the artifact to load.
        validate : bool, optional (default True)
            Whether or not to validate the runtime environment against the artifact
            metadata.
        **kwargs
            Keyword arguments for the handler read function.

        Returns
        -------
        Any
            The artifact object.

        Raises
        ------
        lazyscribe.exception.ArtifactLoadError
            If ``validate`` and runtime environment does not match artifact metadata.
            Or if there is no artifact found with the name provided.
        """
        for artifact in self.artifacts:
            if artifact.name == name:
                # Validate the handler
                if validate:
                    validate_artifact_environment(artifact)
                # Read in the artifact
                mode = "rb" if artifact.binary else "r"
                with self.fs.open(str(self.path / artifact.fname), mode) as buf:
                    out = artifact.read(buf, **kwargs)
                if artifact.output_only:
                    warnings.warn(
                        f"Artifact '{name}' is not the original Python Object",
                        UserWarning,
                        stacklevel=2,
                    )
                break
        else:
            raise ArtifactLoadError(f"No artifact with name {name}")

        return out

    @contextmanager
    def log_test(self, name: str, description: str | None = None) -> Iterator[Test]:
        """Add a test to the experiment using a context handler.

        A test is a specific location for non-global metrics.

        Parameters
        ----------
        name : str
            Name of the test.
        description : str, optional (default None)
            An optional description for the test.

        Yields
        ------
        lazyscribe.test.Test
            The :py:class:`lazyscribe.test.Test` dataclass.
        """
        test = Test(name=name, description=description)

        yield test

        self.last_updated = utcnow()
        self.tests.append(test)

        self.dirty = True

    def to_dict(self) -> dict[str, Any]:
        """Serialize the experiment to a dictionary.

        Returns
        -------
        dict[str, Any]
            The experiment dictionary.
        """
        return asdict(
            self,
            value_serializer=serializer,
            filter=filters.exclude(
                fields(Experiment).dir,
                fields(Experiment).project,
                fields(Experiment).fs,
                fields(Experiment).dirty,
            ),
        )

    def promote_artifact(self, repository: Repository, name: str) -> None:
        """Associate an artifact with a :py:class:`lazyscribe.repository.Repository`.

        The purpose of this method is to move an artifact from an *ephemeral*
        experiment to the versioned repository.

        If the artifact does not exist on disk yet, this function is simply a passthrough
        to :py:meth:`lazyscribe.repository.Repository.log_artifact`. If the artifact does
        exist on disk already, this function will copy the artifact from the experiment
        directory to the repository, increment the version, and call
        :py:meth:`lazyscribe.repository.Repository.save`.

        Parameters
        ----------
        repository : lazyscribe.repository.Repository
            The :py:class:`lazyscribe.repository.Repository` to promote the artifact to.
        name : str
            The artifact to promote.

        Raises
        ------
        lazyscribe.exception.ArtifactLogError
            Raised if the artifact to be promoted is not newer than the latest version available
            in the repository.
            Raised if

            * the artifact ``name`` exists on the filesystem, and
            * the filesystem protocol does not match between the repository and the experiment.
        lazyscribe.exception.ArtifactLoadError
            Raised if there is no artifact with the name ``name`` in the experiment.
        lazyscribe.exception.SaveError
            Raised when writing to the filesystem fails.
        """
        for artifact in self.artifacts:
            if artifact.name == name:
                try:
                    meta_ = repository.get_artifact_metadata(name)
                    if (
                        datetime.strptime(meta_["created_at"], "%Y-%m-%dT%H:%M:%S")
                        >= artifact.created_at
                    ):
                        raise ArtifactLogError(
                            f"Artifact `{name}` is not newer than the latest version available in the repository."
                        ) from None
                    new_handler = evolve(artifact, version=meta_["version"] + 1)
                except ValueError:
                    new_handler = copy(artifact)

                if artifact.dirty:
                    LOG.debug(
                        f"The current value for artifact '{name}' is not on the filesystem."
                    )
                    repository.artifacts.append(new_handler)
                else:
                    # The artifact is on disk, we will have to copy it over
                    curr_path = self.path / artifact.fname
                    if self.fs.protocol != repository.fs.protocol:
                        raise ArtifactLogError(
                            "The repository and the experiment use different filesystems. "
                            f"Please move {curr_path!s} from the experiment filesystem to "
                            "the Repository filesystem and log it manually."
                        )
                    new_path = repository.dir / artifact.name
                    LOG.debug(f"Copying '{curr_path!s}' to '{new_path!s}{os.sep}'")
                    if not self.fs.isdir(f"{new_path!s}{os.sep}"):
                        LOG.debug(f"Creating '{new_path!s}{os.sep}'")
                        self.fs.mkdir(f"{new_path!s}{os.sep}", create_parents=True)
                    self.fs.copy(str(curr_path), f"{new_path!s}{os.sep}")

                    repository.artifacts.append(new_handler)
                    LOG.info(
                        "Calling `save` on the repository since the artifact exists on disk already."
                    )
                    try:
                        repository.save()
                    except SaveError as exc:
                        LOG.info(
                            f"Save failed, deleting '{(new_path / artifact.fname)!s}'..."
                        )
                        self.fs.rm(str(new_path / artifact.fname))
                        del repository.artifacts[-1]

                        raise exc
                break
        else:
            raise ArtifactLoadError(f"No artifact with name {name}")

    def __str__(self) -> str:
        """Shortened string representation."""
        return f"<lazyscribe.experiment.Experiment at {hex(id(self))}>"

    def __gt__(self, other: Experiment) -> bool:
        """Determine whether this experiment is newer than another experiment.

        If the experiments have the same ``slug``, this function will compare using the
        ``last_updated`` attribute. If the ``slug`` is different, this function will use
        the ``created_at`` value.
        """
        if self.slug == other.slug:
            return self.last_updated > other.last_updated
        else:
            return self.created_at > other.created_at

    def __lt__(self, other: Experiment) -> bool:
        """Determine whether this experiment is older than another experiment.

        If the experiments have the same ``slug``, this function will compare using the
        ``last_updated`` attribute. If the ``slug`` is different, this function will use
        the ``created_at`` value.
        """
        if self.slug == other.slug:
            return self.last_updated < other.last_updated
        else:
            return self.created_at < other.created_at

    def __ge__(self, other: Experiment) -> bool:
        """Determine whether this experiment is newer than another experiment.

        If the experiments have the same ``slug``, this function will compare using the
        ``last_updated`` attribute. If the ``slug`` is different, this function will use
        the ``created_at`` value.
        """
        return bool(self == other or self > other)

    def __le__(self, other: Experiment) -> bool:
        """Determine whether this experiment is older than another experiment.

        If the experiments have the same ``slug``, this function will compare using the
        ``last_updated`` attribute. If the ``slug`` is different, this function will use
        the ``created_at`` value.
        """
        return bool(self == other or self < other)


@frozen
class ReadOnlyExperiment(Experiment):
    """Immutable version of an experiment."""

    def __str__(self) -> str:
        """Shortened string representation."""
        return f"<lazyscribe.experiment.ReadOnlyExperiment at {hex(id(self))}>"



"""Project storing and logging."""

from __future__ import annotations

import getpass
import json
import logging
import warnings
from collections.abc import Callable, Iterator
from contextlib import contextmanager
from datetime import datetime
from pathlib import Path
from typing import Any, Literal
from urllib.parse import urlparse

import fsspec

from lazyscribe.artifacts import _get_handler
from lazyscribe.artifacts.base import Artifact
from lazyscribe.exception import ReadOnlyError, SaveError
from lazyscribe.experiment import Experiment, ReadOnlyExperiment
from lazyscribe.linked import LinkedList, merge
from lazyscribe.registry import registry
from lazyscribe.test import ReadOnlyTest, Test

LOG = logging.getLogger(__name__)


class Project:
    """Project class.

    Parameters
    ----------
    fpath : str | pathlib.Path, optional (default "project.json")
        The location of the project file. If no project file exists, this will be the location
        of the output JSON file when ``save`` is called.
    mode : {"r", "a", "w", "w+"}, optional (default "a")
        The mode for opening the project.

        * ``r``: All existing experiments will be loaded as
          :py:class:`lazyscribe.experiment.ReadOnlyExperiment`. No new experiments can be added.
        * ``a``: All existing experiments will be loaded as
          :py:class:`lazyscribe.experiment.ReadOnlyExperiment`. New experiments can be added.
        * ``w``: No existing experiments will be loaded.
        * ``w+``: All experiments will be loaded in editable mode as
          :py:class:`lazyscribe.experiment.Experiment`. Experiments can be added.
    author : str, optional (default None)
        The project author. This author will be used for any new experiments or modifications to
        existing experiments. If not supplied, ``getpass.getuser()`` will be used.
    **storage_options
        Storage options to pass to the filesystem initialization. Will be passed to
        :py:meth:`fsspec.filesystem`.

    Attributes
    ----------
    experiments : list[lazyscribe.experiment.Experiment]
        The list of experiments in the project.

    Raises
    ------
    ValueError
        Raised on invalid ``mode`` value.
    """

    fpath: Path
    mode: Literal["r", "a", "w", "w+"]
    author: str
    storage_options: dict[str, Any]
    experiments: list[Experiment]

    def __init__(
        self,
        fpath: str | Path = "project.json",
        mode: Literal["r", "a", "w", "w+"] = "a",
        author: str | None = None,
        **storage_options: Any,
    ) -> None:
        """Init method."""
        if isinstance(fpath, str):
            parsed = urlparse(fpath)
            self.fpath = Path(parsed.netloc + parsed.path)
            self.protocol = parsed.scheme or "file"
        else:
            self.fpath = fpath
            self.protocol = "file"
        self.storage_options = storage_options

        # If in ``r``, ``a``, or ``w+`` mode, read in the existing project.
        self.experiments: list[Experiment] = []
        self.fs = fsspec.filesystem(self.protocol, **storage_options)

        if mode not in ("r", "a", "w", "w+"):
            raise ValueError("Please provide a valid ``mode`` value.")
        self.mode = mode
        if mode in ("r", "a", "w+") and self.fs.isfile(str(self.fpath)):
            self.load()

        self.author = getpass.getuser() if author is None else author

    def load(self) -> None:
        """Load existing experiments.

        If the project is in read-only or append mode, existing experiments will
        be loaded in read-only mode. If opened in editable mode, existing experiments
        will be loaded in editable mode.
        """
        with self.fs.open(str(self.fpath), "r") as infile:
            data = json.load(infile)
        for idx, entry in enumerate(data):
            data[idx]["created_at"] = datetime.fromisoformat(entry["created_at"])
            data[idx]["last_updated"] = datetime.fromisoformat(entry["last_updated"])

        upstream_projects: dict[str, Project] = {}
        for exp in data:
            dependencies: dict[str, Experiment] = {}
            if "dependencies" in exp:
                deplist = exp.pop("dependencies")
                for dep in deplist:
                    project_name, exp_name = dep.split("|")
                    if project_name in registry:
                        # The upstream experiment refers to the project registry by key
                        project = registry[project_name]
                    elif (project_key_ := registry.search(project_name)) is not None:
                        # The upstream experiment refers to a project path that has been
                        # loaded into the registry
                        project = registry[project_key_]
                    elif project_name in upstream_projects:
                        project = upstream_projects[project_name]
                    else:
                        project = Project(
                            fpath=f"{self.protocol}://{project_name}",
                            mode="r",
                            author=None,
                            **self.storage_options,
                        )
                        project.load()
                        upstream_projects[project_name] = project
                    depexp = project[exp_name]
                    dependencies[depexp.short_slug] = depexp

            tests: list[Test] = []
            if "tests" in exp:
                testlist = exp.pop("tests")
                for test in testlist:
                    if self.mode in ("r", "a"):
                        tests.append(ReadOnlyTest(**test))
                    else:
                        tests.append(Test(**test))

            artifacts: list[Artifact] = []
            if "artifacts" in exp:
                artifactlist = exp.pop("artifacts")
                for artifact in artifactlist:
                    handler_cls = _get_handler(artifact.pop("handler"))
                    created_at = datetime.fromisoformat(artifact.pop("created_at"))
                    artifacts.append(
                        handler_cls.construct(
                            **artifact, created_at=created_at, dirty=False
                        )
                    )

            if self.mode in ("r", "a"):
                self.experiments.append(
                    ReadOnlyExperiment(
                        **exp,
                        project=self.fpath,
                        fs=self.fs,
                        dependencies=dependencies,
                        tests=tests,
                        artifacts=artifacts,
                        dirty=False,
                    )
                )
            else:
                self.experiments.append(
                    Experiment(
                        **exp,
                        project=self.fpath,
                        fs=self.fs,
                        dependencies=dependencies,
                        tests=tests,
                        artifacts=artifacts,
                        dirty=False,
                    )
                )

    def save(self) -> None:
        """Save the project data.

        This includes saving any artifact data.

        Raises
        ------
        lazyscribe.exception.ReadOnlyError
            Raised when trying to save when the project is in read-only mode.
        lazyscribe.exception.SaveError
            Raised when writing to the filesystem fails.
        """
        if self.mode == "r":
            raise ReadOnlyError("Project is in read-only mode.")
        elif self.mode == "w+":
            for exp in self.experiments:
                if exp.dirty:
                    exp.last_updated_by = self.author

        data = list(self)
        with self.fs.transaction:
            try:
                self.fs.makedirs(str(self.fpath.parent), exist_ok=True)
                with self.fs.open(str(self.fpath), "w") as outfile:
                    json.dump(data, outfile, sort_keys=True, indent=4)
            except Exception as exc:
                raise SaveError(
                    f"Unable to save the Project JSON file to {self.fpath!s}"
                ) from exc

            mutable_: list[Experiment] = [
                exp
                for exp in self.experiments
                if not isinstance(exp, ReadOnlyExperiment)
            ]
            for exp in mutable_:
                if not exp.dirty:
                    LOG.debug(f"{exp.slug} has not been updated. Skipping...")
                    continue
                # Write the artifact data
                LOG.info(f"Saving artifacts for {exp.slug}")
                for artifact in exp.artifacts:
                    fmode = "wb" if artifact.binary else "w"
                    fpath = exp.path / artifact.fname
                    if not artifact.dirty:
                        LOG.debug(f"Artifact '{artifact.name}' has not been updated")
                        continue

                    try:
                        self.fs.makedirs(str(exp.path), exist_ok=True)
                        LOG.debug(f"Saving '{artifact.name}' to {fpath!s}...")
                        with self.fs.open(str(fpath), fmode) as buf:
                            artifact.write(
                                artifact.value, buf, **artifact.writer_kwargs
                            )
                    except Exception as exc:
                        raise SaveError(
                            f"Unable to write '{artifact.name}' to '{fpath!s}'"
                        ) from exc

                    # Reset the `dirty` flag since we have the updated artifact on disk
                    artifact.dirty = False
                    if artifact.output_only:
                        warnings.warn(
                            f"Artifact '{artifact.name}' is added. It is not meant to be read back as Python Object",
                            UserWarning,
                            stacklevel=2,
                        )

                exp.dirty = False

    def merge(self, other: Project) -> Project:
        """Merge two projects.

        The new project will inherit the current project ``fpath``,
        ``author``, and ``mode``.

        For details on the merging process, see :ref:`here <Project Updating>`.

        Returns
        -------
        lazyscribe.project.Project
            A new project.
        """
        # Create linked lists of experiments
        cexp = LinkedList.from_list(self.experiments)
        oexp = LinkedList.from_list(other.experiments)
        # Get the merged list of experiments
        merged = merge(cexp.head, oexp.head).to_list()  # type: ignore[arg-type]
        # De-dupe the merged list based on slug
        slugs: list[str] = [exp.slug for exp in merged]

        new = Project(
            fpath=self.fpath,
            mode=self.mode,
            author=self.author,
            **self.storage_options,
        )
        new.experiments = [
            val for idx, val in enumerate(merged) if val.slug not in slugs[idx + 1 :]
        ]

        return new

    def append(self, other: Experiment) -> None:
        """Append an experiment to the project.

        For details on the merging process, see :ref:`here <Project Appending>`.

        Parameters
        ----------
        other : lazyscribe.experiment.Experiment
            The experiment to add.

        Raises
        ------
        lazyscribe.exception.ReadOnlyError
            Raised when trying to log a new experiment when the project is in
            read-only mode.
        """
        if self.mode == "r":
            raise ReadOnlyError("Project is in read-only mode.")
        self.experiments.append(other)

    @contextmanager
    def log(self, name: str) -> Iterator[Experiment]:
        """Log an experiment to the project.

        Parameters
        ----------
        name : str
            The name of the experiment.

        Yields
        ------
        lazyscribe.experiment.Experiment
            A new :py:class:`lazyscribe.experiment.Experiment` object.

        Raises
        ------
        lazyscribe.exception.ReadOnlyError
            Raised when trying to log a new experiment when the project is in
            read-only mode.
        """
        if self.mode == "r":
            raise ReadOnlyError("Project is in read-only mode.")
        experiment = Experiment(
            name=name, project=self.fpath, fs=self.fs, author=self.author, dirty=True
        )

        yield experiment

        self.append(experiment)

    def filter(self, func: Callable[[Experiment], bool]) -> Iterator[Experiment]:
        """Filter the experiments in the project.

        Parameters
        ----------
        func : Callable[[lazyscribe.experiment.Experiment], bool]
            A callable that takes in a :py:class:`lazyscribe.experiment.Experiment` object
            and returns a boolean indicating whether or not it passes the filter.

        Yields
        ------
        lazyscribe.experiment.Experiment
            An experiment.
        """
        for exp in self.experiments:
            if func(exp):
                yield exp

    def __contains__(self, item: str) -> bool:
        """Check if the project contains an experiment with the given slug or short slug.

        Parameters
        ----------
        item : str
            The slug or short slug for the experiment.

        Returns
        -------
        bool
            Whether the item exists.
        """
        for exp in self.experiments:
            if exp.slug == item or exp.short_slug == item:
                out = True
                break
        else:
            out = False

        return out

    def __getitem__(self, arg: str) -> Experiment:
        """Use brackets to retrieve an experiment by slug.

        Parameters
        ----------
        arg : str
            The slug or short slug for the experiment.

            .. note::

                If you have multiple experiments with the same short slug, this notation
                will retrieve the first one added to the project.

        Raises
        ------
        KeyError
            Raised if the slug does not exist.
        """
        for exp in self.experiments[
            ::-1
        ]:  # reverse list to get the latest saved experiment
            if exp.slug == arg or exp.short_slug == arg:
                out = exp
                break
        else:
            raise KeyError(f"No experiment with slug {arg}")

        # If short slug is used, return the latest experiment saved with that slug, but may not correspond to the latest "last_updated" timestamp if it was manually set.
        short_slug_timestamp: list[datetime] = [
            exp.last_updated for exp in self.experiments if exp.short_slug == arg
        ]
        if len(short_slug_timestamp) > 0:
            latest_datetime = max(short_slug_timestamp)
            if out.last_updated != latest_datetime:
                LOG.warning(
                    f"Returning experiment last saved, with ``last_updated`` manually set as {exp.last_updated}. \
                            The latest ``last_updated`` timestamp is {latest_datetime}."
                )

        return out

    def __iter__(self) -> Iterator[dict[str, Any]]:
        """Iterate through each experiment and return the dictionary.

        Yields
        ------
        dict[str, Any]
            An experiment dict.
        """
        for exp in self.experiments:
            yield exp.to_dict()



"""Custom exceptions for lazyscribe."""


class LazyscribeError(Exception):
    """Base exception for lazyscribe errors."""


class ReadOnlyError(LazyscribeError):
    """Raised when a project or repository is opened in read-only mode and write operations are tried."""


class SaveError(LazyscribeError):
    """Raised when a project or repository is unable to save objects to the filesystem."""


class ArtifactError(LazyscribeError):
    """Base exception for artifact errors."""


class ArtifactLogError(ArtifactError):
    """Raised when an artifact cannot be logged."""


class ArtifactLoadError(ArtifactError):
    """Raised when an artifact cannot be loaded."""


class VersionNotFoundError(LazyscribeError):
    """Raised when the version cannot be found."""


class InvalidVersionError(LazyscribeError):
    """Raised when an invalid version is provided."""



"""Project registry.

This module defines a class used to make inter-filesystem work
easier.
"""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from attrs import Factory, define

if TYPE_CHECKING:
    from lazyscribe import Project


@define
class Registry:
    """A registry of existing projects to reference by name.

    Parameters
    ----------
    projects : dict[str, lazyscribe.project.Project]
        A dictionary with a key for each available project.
    """

    projects: dict[str, Project] = Factory(factory=lambda: {})

    def add_project(self, name: str, project: Project) -> None:
        """Add a project to the registry.

        Parameters
        ----------
        name : str
            The key for the project.
        project : lazyscribe.project.Project
            The project object itself.
        """
        self.projects[name] = project

    def search(self, fpath: Path) -> str | None:
        """Find a project from the JSON path.

        This method helps identify a project key in the repository based
        on the location of the JSON file.

        Parameters
        ----------
        fpath : pathlib.Path
            Path to the project JSON.

        Returns
        -------
        str | None
            If it exists, the key reference for the project in the registry.
        """
        out: str | None = None
        for name, project in self.projects.items():
            if project.fpath == fpath:
                out = name
                break

        return out

    def __getitem__(self, arg: str) -> Project:
        """Retrieve a project.

        Parameters
        ----------
        arg : str
            The name of a project.

        Returns
        -------
        lazyscribe.project.Project
            The project object.

        Raises
        ------
        KeyError
            Raised if the project name does not exists in the registry.
        """
        try:
            return self.projects[arg]
        except KeyError as exc:
            msg = f"No project with the name '{arg}' in the registry"
            raise KeyError(msg) from exc

    def __contains__(self, key: str) -> bool:
        """Check if a project is in the registry.

        Parameters
        ----------
        key : str
            The name of the project.

        Returns
        -------
        bool
            Whether the project name exists in the registry.
        """
        return key in self.projects


registry = Registry()



"""Pickle-based handler for binary artifacts."""

from __future__ import annotations

import logging
import pickle
import sys
from datetime import datetime
from io import IOBase
from typing import Any, ClassVar

from attrs import define, field
from slugify import slugify

from lazyscribe._utils import utcnow
from lazyscribe.artifacts.base import Artifact

LOG = logging.getLogger(__name__)


@define(auto_attribs=True)
class PickleArtifact(Artifact):
    """Pickle-based serialization for python objects.

    .. important::

        This class is not meant to be initialized directly. Please use the ``construct``
        method.

    .. note::

        For the attributes documentation, please see the "Attributes" section of
        :py:class:`lazyscribe.artifacts.base.Artifact`.

    Attributes
    ----------
    alias : str = "pickle"
    suffix : str = "pkl"
    binary : bool = False
    output_only : bool = False

    python_version : str
        Minor Python version (e.g. ``"3.10"``).
    """

    alias: ClassVar[str] = "pickle"
    suffix: ClassVar[str] = "pkl"
    binary: ClassVar[bool] = True
    output_only: ClassVar[bool] = False
    python_version: str = field()

    @classmethod
    def construct(
        cls,
        name: str,
        value: Any = None,
        fname: str | None = None,
        created_at: datetime | None = None,
        expiry: datetime | None = None,
        writer_kwargs: dict[str, Any] | None = None,
        version: int = 0,
        dirty: bool = True,
        **kwargs: Any,
    ) -> PickleArtifact:
        """Construct the handler class.

        Parameters
        ----------
        name : str
            The name of the artifact.
        value : Any, optional (default None)
            The value for the artifact. The default value of ``None`` is used when
            an experiment is loaded from the project JSON.
        fname : str, optional (default None)
            The filename for the artifact. If set to ``None`` or not provided, it will be derived from
            the name of the artifact and the suffix for the class.
        created_at : datetime.datetime, optional (default ``lazyscribe._utils.utcnow()``)
            When the artifact was created.
        expiry : datetime.datetime, optional (default None)
            When the artifact expired.
        writer_kwargs : dict[str, Any], optional (default {})
            Keyword arguments for writing an artifact to the filesystem. Provided when an artifact
            is logged to an experiment.
        version : int, optional (default 0)
            Integer version to be used for versioning artifacts.
        dirty : bool, optional (default True)
            Whether or not this artifact should be saved when :py:meth:`lazyscribe.project.Project.save`
            or :py:meth:`lazyscribe.repository.Repository.save` is called. This decision is based
            on whether the artifact is new or has been updated.
        python_version : str, optional
            Minor Python version (e.g. ``"3.10"``).

        Returns
        -------
        JSONArtifact
            The artifact.
        """
        python_version = kwargs.get("python_version") or ".".join(
            str(i) for i in sys.version_info[:2]
        )
        created_at = created_at or utcnow()
        return cls(
            name=name,
            value=value,
            fname=fname
            or f"{slugify(name)}-{slugify(created_at.strftime('%Y%m%d%H%M%S'))}.{cls.suffix}",
            created_at=created_at,
            expiry=expiry,
            writer_kwargs=writer_kwargs or {},
            version=version,
            dirty=dirty,
            python_version=python_version,
        )

    @classmethod
    def read(cls, buf: IOBase, **kwargs: Any) -> Any:
        """Read in the file.

        Parameters
        ----------
        buf : file-like object
            The buffer from a ``fsspec`` filesystem.
        **kwargs
            Keyword arguments for :py:meth:`pickle.load`.

        Returns
        -------
        Any
            The deserialized Python object.
        """
        return pickle.load(buf, **kwargs)

    @classmethod
    def write(cls, obj: Any, buf: IOBase, **kwargs: Any) -> None:
        """Write the content to a pickle file.

        Parameters
        ----------
        obj : object
            The serializable object.
        buf : file-like object
            The buffer from a ``fsspec`` filesystem.
        **kwargs
            Keyword arguments for :py:meth:`pickle.dump`.
        """
        pickle.dump(obj, buf, **kwargs)



"""Repository storing and logging."""

from __future__ import annotations

import copy
import difflib
import json
import logging
import warnings
from bisect import bisect
from collections.abc import Iterator
from datetime import datetime
from operator import attrgetter
from pathlib import Path
from typing import Any, Literal
from urllib.parse import urlparse

import fsspec

from lazyscribe._utils import serialize_artifacts, utcnow, validate_artifact_environment
from lazyscribe.artifacts import _get_handler
from lazyscribe.artifacts.base import Artifact
from lazyscribe.exception import (
    ArtifactLoadError,
    InvalidVersionError,
    ReadOnlyError,
    SaveError,
    VersionNotFoundError,
)

LOG = logging.getLogger(__name__)


class Repository:
    """Repository class for holding versioned artifacts.

    Parameters
    ----------
    fpath : str | Path, optional (default "repository.json")
        The location of the repository file. If no repository file exists, this will be the location
        of the output JSON file when ``save`` is called.
    mode : {"r", "a", "w", "w+"}, optional (default "w+")
        The mode for opening the repository.

        * ``r``: All artifacts will be loaded. No new artifacts can be logged.
        * ``w``: No existing artifacts will be loaded. Artifacts can be added.
        * ``w+``: All artifacts will be loaded. New artifacts can be added.
    **storage_options
        Storage options to pass to the filesystem initialization. Will be passed to
        :py:meth:`fsspec.filesystem`.

    Attributes
    ----------
    artifacts : list[lazyscribe.artifact.Artifact]
        The list of artifacts in the repository.
    """

    fpath: Path
    mode: Literal["r", "w", "w+"]
    storage_options: dict[str, Any]
    artifacts: list[Artifact]

    def __init__(
        self,
        fpath: str | Path = "repository.json",
        mode: Literal["r", "w", "w+"] = "w+",
        **storage_options: Any,
    ) -> None:
        """Init method.

        Raises
        ------
        ValueError
            Raised on invalid ``mode`` value.
        """
        if isinstance(fpath, str):
            parsed = urlparse(fpath)
            self.fpath = Path(parsed.netloc + parsed.path)
            self.protocol = parsed.scheme or "file"
        else:
            self.fpath = fpath
            self.protocol = "file"
        self.dir = self.fpath.parent
        self.storage_options = storage_options

        self.artifacts: list[Artifact] = []
        self.fs = fsspec.filesystem(self.protocol, **storage_options)

        if mode not in ("r", "w", "w+"):
            raise ValueError("Please provide a valid ``mode`` value.")
        self.mode = mode
        if mode in ("r", "w+") and self.fs.isfile(str(self.fpath)):
            self.load()

    def load(self) -> None:
        """Load existing artifacts."""
        with self.fs.open(str(self.fpath), "r") as infile:
            data = json.load(infile)

        artifacts: list[Artifact] = []
        for artifact in data:
            handler_cls = _get_handler(artifact.pop("handler"))
            created_at = datetime.fromisoformat(artifact.pop("created_at"))
            artifacts.append(
                handler_cls.construct(**artifact, created_at=created_at, dirty=False)
            )
        self.artifacts = artifacts

    def log_artifact(
        self,
        name: str,
        value: Any,
        handler: str,
        fname: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Log an artifact to the repository.

        This method associates an artifact with the repository, but the artifact will
        not be written until :py:meth:`lazyscribe.repository.Repository.save` is called.

        Parameters
        ----------
        name : str
            The name of the artifact.
        value : Any
            The object to persist to the filesystem.
        handler : str
            The name of the handler to use for the object.
        fname : str, optional (default None)
            The filename for the artifact. If set to ``None`` or not provided, it will be derived
            from the name of the artifact and the builtin suffix for each handler.
        **kwargs
            Keyword arguments for the write function of the handler.

        Raises
        ------
        lazyscribe.exception.ReadOnlyError
            If repository is in read-only mode.
        """
        if self.mode == "r":
            raise ReadOnlyError("Repository is in read-only mode.")
        # Retrieve and construct the handler
        self.last_updated = utcnow()
        artifacts_matching_name: list[Artifact] = [
            art for art in self.artifacts if art.name == name
        ]
        version = (
            max(art.version for art in artifacts_matching_name) + 1
            if artifacts_matching_name
            else 0
        )
        handler_cls = _get_handler(handler)
        artifact_handler = handler_cls.construct(
            name=name,
            value=value,
            fname=fname,
            created_at=self.last_updated,
            writer_kwargs=kwargs,
            version=version,
        )
        self.artifacts.append(artifact_handler)
        if handler_cls.output_only:
            warnings.warn(
                f"Artifact '{name}' is added. It is not meant to be read back as Python Object",
                UserWarning,
                stacklevel=2,
            )

    def load_artifact(
        self,
        name: str,
        validate: bool = True,
        version: datetime | str | int | None = None,
        match: Literal["asof", "exact"] = "exact",
        **kwargs: Any,
    ) -> Any:
        """Load a single artifact.

        Parameters
        ----------
        name : str
            The name of the artifact to load.
        validate : bool, optional (default True)
            Whether or not to validate the runtime environment against the artifact
            metadata.
        version : datetime.datetime | str | int, optional (default None)
            The version of the artifact to load.
            Can be provided as a datetime corresponding to the ``created_at`` field,
            a string corresponding to the ``created_at`` field in the format ``"%Y-%m-%dT%H:%M:%S"``
            (e.g. ``"2025-01-25T12:36:22"``), or an integer version.
            If set to ``None`` or not provided, defaults to the most recent version.
        match : {"asof", "exact"}, optional (default "exact")
            Matching logic. Only relevant for ``str`` and ``datetime.datetime`` values for
            ``version``. ``exact`` will provide an artifact with the exact ``created_at``
            value provided. ``asof`` will provide the most recent version as of the
            ``version`` value.

            When using an ``asof`` search, only artifacts with no expiry or an ``expiry``
            datetime set _after_ the provided ``version`` value will be included. When
            ``version=None``, any artifacts that are expired as of
            :py:meth:`lazyscribe._utils.utcnow` will be excluded.
        **kwargs
            Keyword arguments for the handler read function.

        Returns
        -------
        Any
            The artifact object.

        Raises
        ------
        ValueError
            Raised on invalid ``match`` value.
            Raised if no valid artifact was found.
        lazyscribe.exception.ArtifactLoadError
            Raised if ``validate`` and runtime environment does not match artifact metadata.
        """
        # Search for the artifact
        artifact = self._search_artifact_versions(
            name=name, version=version, match=match
        )
        if validate:
            validate_artifact_environment(artifact)

        # Read in the artifact
        mode = "rb" if artifact.binary else "r"
        with self.fs.open(str(self.dir / artifact.name / artifact.fname), mode) as buf:
            out = artifact.read(buf, **kwargs)
        if artifact.output_only:
            warnings.warn(
                f"Artifact '{name}' is not the original Python Object",
                UserWarning,
                stacklevel=2,
            )

        return out

    def set_artifact_expiry(
        self,
        name: str,
        version: datetime | str | int | None = None,
        match: Literal["asof", "exact"] = "exact",
        expiry: datetime | str | None = None,
    ) -> None:
        """Set the expiry date for a given artifact.

        Parameters
        ----------
        name : str
            The name of the artifact.
        version : datetime.datetime | str | int, optional (default None)
            The version of the artifact to load.
            Can be provided as a datetime corresponding to the ``created_at`` field,
            a string corresponding to the ``created_at`` field in the format ``"%Y-%m-%dT%H:%M:%S"``
            (e.g. ``"2025-01-25T12:36:22"``), or an integer version.
            If set to ``None`` or not provided, defaults to the most recent version.
        match : {"asof", "exact"}, optional (default "exact")
            Matching logic. Only relevant for ``str`` and ``datetime.datetime`` values for
            ``version``. ``exact`` will provide an artifact with the exact ``created_at``
            value provided. ``asof`` will provide the most recent version as of the
            ``version`` value.
        expiry : datetime.datetime | str, optional (default None)
            The expiry datetime for the artifact version.

        Raises
        ------
        ValueError
            Raised if the value for ``expiry`` cannot be coerced to a datetime.
        """
        if self.mode == "r":
            raise ReadOnlyError("Repository is in read-only mode.")

        # Search for the artifact
        artifact = self._search_artifact_versions(
            name=name, version=version, match=match
        )
        match expiry:
            case datetime():
                artifact.expiry = expiry
            case str():
                artifact.expiry = datetime.strptime(expiry, "%Y-%m-%dT%H:%M:%S")
            case None:
                artifact.expiry = utcnow()
            case _:
                raise ValueError("Value '%s' cannot be coerced to a datetime", expiry)

    def get_artifact_metadata(
        self,
        name: str,
        version: datetime | str | int | None = None,
        match: Literal["asof", "exact"] = "exact",
    ) -> dict[str, Any]:
        """Retrieve the metadata for an artifact.

        Parameters
        ----------
        name : str
            The name of the artifact to load.
        version : datetime.datetime | str | int, optional (default None)
            The version of the artifact to load.
            Can be provided as a datetime corresponding to the ``created_at`` field,
            a string corresponding to the ``created_at`` field in the format ``"%Y-%m-%dT%H:%M:%S"``
            (e.g. ``"2025-01-25T12:36:22"``), or an integer version.
            If set to ``None`` or not provided, defaults to the most recent version.
        match : {"asof", "exact"}, optional (default "exact")
            Matching logic. Only relevant for ``str`` and ``datetime.datetime`` values for
            ``version``. ``exact`` will provide an artifact with the exact ``created_at``
            value provided. ``asof`` will provide the most recent version as of the
            ``version`` value.

        Returns
        -------
        dict[str, Any]
            The artifact metadata.

        Raises
        ------
        ValueError
            Raised on invalid ``match`` value.
            Raised if no valid artifact was found.
        """
        artifact = self._search_artifact_versions(
            name=name, version=version, match=match
        )

        return next(serialize_artifacts([artifact]))

    def get_version_diff(
        self,
        name: str,
        version: datetime
        | str
        | int
        | tuple[datetime | str | int, datetime | str | int],
        match: Literal["asof", "exact"] = "exact",
    ) -> str:
        """Generate the unified diff between versions of the same artifact.

        Parameters
        ----------
        name : str
            The name of the artifact to compare.
        versions : datetime | str | int | tuple[datetime | str | int, datetime | str | int]
            The versions to compare. If a single version is provided, the artifact will
            be compared to the latest available artifact. A tuple specifies the two versions
            to compare.
        match : {"asof", "exact"}, optional (default "exact")
            Matching logic. Only relevant for ``str`` and ``datetime.datetime`` values.
            ``exact`` will provide an artifact with the exact ``created_at``
            value provided. ``asof`` will provide the most recent version as of the
            ``version`` value.

        Raises
        ------
        lazyscribe.exception.ArtifactLoadError
            Raised if the artifact does not exist on the filesystem yet.
        ValueError
            Raised if the provided artifact(s) represent binary files.

        Returns
        -------
        str
            Concatenated output from :py:meth:`difflib.unified_diff`.
        """
        if not isinstance(version, tuple):
            # Retrieve the specified and the latest version
            versions = [
                self._search_artifact_versions(name, version, match),
                self._search_artifact_versions(name, None, match),
            ]
        else:
            versions = sorted(
                [self._search_artifact_versions(name, ver, match) for ver in version],
                key=lambda x: x.created_at,
            )

        if versions[0] == versions[1]:
            LOG.warning(
                f"Only version '{versions[0].version!s}' was supplied for comparison"
            )

        raw_version_data = []
        for art in versions:
            if art.dirty:
                msg = (
                    "Artifact version not found on the filesystem. Please call"
                    " `Repository.save` before calling this method."
                )
                raise ArtifactLoadError(msg)
            if art.binary:
                msg = (
                    f"Version {version} of '{name}' is written to the filesystem "
                    f"using binary handler '{art.alias}'. Binary file formats cannot"
                    "be compared using diffs."
                )
                raise ValueError(msg)
            with self.fs.open(str(self.dir / art.name / art.fname), "r") as buf:
                raw_version_data.append(buf.read().splitlines())

        return "\n".join(list(difflib.unified_diff(*raw_version_data)))

    def save(self) -> None:
        """Save the repository data.

        This includes saving any artifact data.

        Raises
        ------
        lazyscribe.exception.ReadOnlyError
            Raised when trying to save when the project is in read-only mode.
        lazyscribe.exception.SaveError
            Raised when writing to the filesystem fails.
        """
        if self.mode == "r":
            raise ReadOnlyError("Repository is in read-only mode.")

        data = list(self)
        with self.fs.transaction:
            try:
                self.fs.makedirs(str(self.fpath.parent), exist_ok=True)
                with self.fs.open(str(self.fpath), "w") as outfile:
                    json.dump(data, outfile, sort_keys=True, indent=4)
            except Exception as exc:
                raise SaveError(
                    f"Unable to save the Repository JSON file to {self.fpath!s}"
                ) from exc

            for artifact in self.artifacts:
                # Write the artifact data
                fmode = "wb" if artifact.binary else "w"
                artifact_dir = self.dir / artifact.name
                fpath = artifact_dir / artifact.fname
                if not artifact.dirty:
                    LOG.debug(
                        f"Artifact {artifact.name} v{artifact.version} already exists and has not been updated"
                    )
                    continue

                try:
                    self.fs.makedirs(str(artifact_dir), exist_ok=True)
                    LOG.debug(f"Saving '{artifact.name}' to {fpath!s}...")
                    with self.fs.open(str(fpath), fmode) as buf:
                        artifact.write(artifact.value, buf, **artifact.writer_kwargs)
                except Exception as exc:
                    raise SaveError(
                        f"Unable to write '{artifact.name}' to '{fpath!s}'"
                    ) from exc

                # Reset the `dirty` flag since we have the updated artifact on disk
                artifact.dirty = False
                if artifact.output_only:
                    warnings.warn(
                        f"Artifact '{artifact.name}' is added. It is not meant to be read back as Python Object",
                        UserWarning,
                        stacklevel=2,
                    )

    def filter(
        self, version: datetime | str | list[tuple[str, datetime | str | int]]
    ) -> Repository:
        """Filter a repository.

        This method returns a new, read-only object with a subset of the input artifacts.
        Use this method to truncate a repository to a collection of artifacts relevant to
        a given use case.

        Parameters
        ----------
        version : datetime.datetime | str | list[tuple[str, datetime.datetime | str | int]]
            The version corresponding to the output version of each artifact. If a datetime
            or string is provided, this method will do an ``asof`` search for each artifact.

            If a list is provided, it will be treated as a list of exact versions to load.

        Returns
        -------
        lazyscribe.repository.Repository
            A read-only copy of the existing repository with one version per artifact.

        Raises
        ------
        RuntimeError
            Raised if the current repository object has artifacts that have not been saved
            to the filesystem.
        """
        if self.mode != "r":
            raise RuntimeError("Repository must be in read-only mode for filtering.")
        if any(art.dirty for art in self.artifacts):
            raise RuntimeError(
                "At least one artifact has changed since it was last saved. Please save your "
                "repository and re-open it in read-only mode before filtering."
            )

        new_ = copy.copy(self)
        all_artifacts_ = {art.name for art in new_.artifacts}

        new_.artifacts = []
        match version:
            case datetime() | str():
                for art in all_artifacts_:
                    try:
                        latest = self._search_artifact_versions(
                            name=art, version=version, match="asof"
                        )
                        new_.artifacts.append(latest)
                    except VersionNotFoundError:
                        LOG.warning(
                            f"Artifact '{art}' does not have a version that predates {version!s}"
                        )
            case list():
                for art, ver in version:
                    new_.artifacts.append(
                        self._search_artifact_versions(
                            name=art, version=ver, match="exact"
                        )
                    )

        return new_

    def _search_artifact_versions(
        self,
        name: str,
        version: datetime | str | int | None = None,
        match: Literal["asof", "exact"] = "exact",
    ) -> Artifact:
        """Search for an artifact based on name and version.

        Parameters
        ----------
        name : str
            The name of the artifact to load.
        version : datetime.datetime | str | int, optional (default None)
            The version of the artifact to load.
            Can be provided as a datetime corresponding to the ``created_at`` field,
            a string corresponding to the ``created_at`` field in the format ``"%Y-%m-%dT%H:%M:%S"``
            (e.g. ``"2025-01-25T12:36:22"``), or an integer version.
            If set to ``None`` or not provided, defaults to the most recent version.
        match : {"asof", "exact"}, optional (default "exact")
            Matching logic. Only relevant for ``str`` and ``datetime.datetime`` values for
            ``version``. ``exact`` will provide an artifact with the exact ``created_at``
            value provided. ``asof`` will provide the most recent version as of the
            ``version`` value.

        Raises
        ------
        ValueError
            Raised if there are no artifacts with the provided name.
            Raised if ``match`` is an invalid value.
        VersionNotFoundError
            Raised if the version cannot be found.
        InvalidVersionError
            Raised if the version cannot be coerced. Only relevant for string values.
        """
        artifacts_matching_name = sorted(
            [art for art in self.artifacts if art.name == name],
            key=lambda x: x.created_at,
        )
        if not artifacts_matching_name:
            raise ValueError(f"No artifact with name {name}")
        try:
            version = (
                datetime.strptime(version, "%Y-%m-%dT%H:%M:%S")
                if isinstance(version, str)
                else version
            )
        except ValueError as exc:
            msg = f"Invalid version identifier provided. {version} is not in the format YYYY-MM-DDTHH:MM:SS"
            raise InvalidVersionError(msg) from exc

        match (match, version):
            case (_, None):
                today = utcnow()
                best_before_ = [
                    art
                    for art in artifacts_matching_name
                    if art.expiry is None or today < art.expiry
                ]
                try:
                    artifact = best_before_[-1]
                except IndexError as exc:
                    msg = (
                        f"Using {today!s} as a reference, all artifacts with "
                        f"the name '{name}' are expired."
                    )
                    raise VersionNotFoundError(msg) from exc
            case ("exact", datetime()):
                try:
                    artifact = next(
                        art
                        for art in artifacts_matching_name
                        if art.created_at == version
                    )
                except StopIteration:
                    raise VersionNotFoundError(
                        f"No artifact named {name} with version {version}"
                    ) from None
            case ("asof", datetime()):
                LOG.info(
                    f"Searching for the latest version of '{name}' as of {version!s}..."
                )

                eligible_idx_ = bisect(
                    artifacts_matching_name, version, key=attrgetter("created_at")
                )
                eligible_artifacts_ = artifacts_matching_name[:eligible_idx_]
                if len(eligible_artifacts_) == 0:
                    msg = (
                        f"Version {version!s} predates the earliest version "
                        f"{artifacts_matching_name[0].created_at!s}."
                    )
                    raise VersionNotFoundError(msg) from None

                for art in eligible_artifacts_[::-1]:
                    if art.expiry is None or version < art.expiry:
                        artifact = art
                        break
                else:
                    msg = (
                        f"The only available artifacts with name '{name}' as of "
                        f"{version!s} are expired."
                    )
                    raise VersionNotFoundError(msg) from None
                LOG.info(
                    f"Found version {artifact.version} (created {artifact.created_at!s})"
                )
            case ("exact", int()):
                try:
                    # Integer version is 0-indexed
                    artifact = next(
                        art for art in artifacts_matching_name if art.version == version
                    )
                except StopIteration:
                    raise VersionNotFoundError(
                        f"No artifact named {name} with version {version}"
                    ) from None
            case _:
                raise ValueError(
                    "Please provide ``exact`` or ``asof`` for match. ``asof`` is only "
                    "valid for str/datetime versions."
                )

        return artifact

    def __contains__(self, item: str) -> bool:
        """Check if the repository contains an artifact with the given slug or short slug."""
        return any(art.name == item for art in self.artifacts)

    def __getitem__(self, arg: str) -> Artifact:
        """Use brackets to retrieve an artifact by slug.

        Parameters
        ----------
        arg : str
            The slug or short slug for the artifact.

            .. note::

                If you have multiple artifacts with the same short slug, this notation
                will retrieve the first one added to the repository.

        Raises
        ------
        KeyError
            Raised if the slug does not exist.
        """
        for art in self.artifacts:
            if art.name == arg:
                out = art
                break
        else:
            raise KeyError(f"No artifact with name {arg}")

        return out

    def __iter__(self) -> Iterator[dict[str, Any]]:
        """Iterate through each artifact and return the dictionary."""
        yield from serialize_artifacts(self.artifacts)



"""Artifact handler for JSON-serializable objects."""

from __future__ import annotations

from datetime import datetime
from io import IOBase
from json import dump, load
from typing import Any, ClassVar

from attrs import define
from slugify import slugify

from lazyscribe._utils import utcnow
from lazyscribe.artifacts.base import Artifact


@define(auto_attribs=True)
class JSONArtifact(Artifact):
    """Handler for JSON-serializable objects.

    .. important::

        This class is not meant to be initialized directly. Please use the ``construct``
        method.

    .. note::

        For the attributes documentation, see also "Attributes" of :py:class:`lazyscribe.artifacts.base.Artifact`.

    Attributes
    ----------
    alias : str = "json"
    suffix : str = "json"
    binary : bool = False
    output_only : bool = False

    python_version : str
        Minor Python version (e.g. ``"3.10"``).
    """

    alias: ClassVar[str] = "json"
    suffix: ClassVar[str] = "json"
    binary: ClassVar[bool] = False
    output_only: ClassVar[bool] = False

    @classmethod
    def construct(
        cls,
        name: str,
        value: Any = None,
        fname: str | None = None,
        created_at: datetime | None = None,
        expiry: datetime | None = None,
        writer_kwargs: dict[str, Any] | None = None,
        version: int = 0,
        dirty: bool = True,
        **kwargs: Any,
    ) -> JSONArtifact:
        """Construct the handler class.

        Parameters
        ----------
        name : str
            The name of the artifact.
        value : Any, optional (default None)
            The value for the artifact. The default value of ``None`` is used when
            an experiment is loaded from the project JSON.
        fname : str, optional (default None)
            The filename for the artifact. If set to ``None`` or not provided, it will be derived from
            the name of the artifact and the suffix for the class.
        created_at : datetime.datetime, optional (default ``lazyscribe._utils.utcnow()``)
            When the artifact was created.
        expiry : datetime.datetime, optional (default None)
            When the artifact expired.
        writer_kwargs : dict[str, Any], optional (default {})
            Keyword arguments for writing an artifact to the filesystem. Provided when an artifact
            is logged to an experiment.
        version : int, optional (default 0)
            Integer version to be used for versioning artifacts.
        dirty : bool, optional (default True)
            Whether or not this artifact should be saved when :py:meth:`lazyscribe.project.Project.save`
            or :py:meth:`lazyscribe.repository.Repository.save` is called. This decision is based
            on whether the artifact is new or has been updated.

        Returns
        -------
        JSONArtifact
            The artifact.
        """
        created_at = created_at or utcnow()
        return cls(
            name=name,
            value=value,
            fname=fname
            or f"{slugify(name)}-{slugify(created_at.strftime('%Y%m%d%H%M%S'))}.{cls.suffix}",
            created_at=created_at,
            expiry=expiry,
            writer_kwargs=writer_kwargs or {},
            version=version,
            dirty=dirty,
        )

    @classmethod
    def read(cls, buf: IOBase, **kwargs: Any) -> Any:
        """Read in the JSON file.

        Parameters
        ----------
        buf : file-like object
            The buffer from a ``fsspec`` filesystem.
        **kwargs
            Keyword arguments for :py:meth:`json.load`

        Returns
        -------
        Any
            The deserialized JSON file.
        """
        return load(buf, **kwargs)

    @classmethod
    def write(cls, obj: Any, buf: IOBase, **kwargs: Any) -> None:
        """Write the content to a JSON file.

        Parameters
        ----------
        obj : object
            The JSON-serializable object.
        buf : file-like object
            The buffer from a ``fsspec`` filesystem.
        **kwargs
            Keyword arguments for :py:meth:`json.dump`.
        """
        dump(obj, buf, **kwargs)



"""Import the handlers."""

from importlib.metadata import entry_points

from lazyscribe.artifacts.base import Artifact

__all__: list[str] = ["_get_handler"]


def _get_handler(alias: str) -> type[Artifact]:
    """Retrieve a specific handler based on the alias.

    Parameters
    ----------
    alias : str
        The alias for the handler.

    Returns
    -------
    type[lazyscribe.artifacts.base.Artifact]
        The artifact handler class.
    """
    entry = entry_points(group="lazyscribe.artifact_type")

    for full_artifact_class in entry:  # search through entrypoints first
        if full_artifact_class.name == alias:
            try:
                mod = full_artifact_class.load()
            except ImportError as imp:
                raise ImportError(
                    f"Unable to import handler for {alias} through entry points"
                ) from imp

            if not isinstance(mod, type):
                raise TypeError(f"{full_artifact_class} is not a class")

            break

    else:
        for obj in Artifact.__subclasses__():  # search through experiment subclasses
            if obj.alias == alias:
                mod = obj
                break

        # no handler found in both entrypoints or subclass
        else:
            raise ValueError(
                f"No handler available with the name {alias} in `artifact_type` group."
            )

    return mod  # type: ignore
